#!/usr/bin/env python3
import random
import string
import sys
import math
import numpy as np
import timeit
import matplotlib.pyplot as plt


sys.setrecursionlimit(6000)  # nastavi maximalni hloubku rekurze na 6000

# Vasim ukolem bude naimplementovat ruzne algoritmy v ruznych tridach
# casove narocnosti. Testy obsahuji mereni rychlosti, takze si 
# pro ruzne casove tridy muzete udelat predstavu, jak dlouho bezi.
# Doba behu neni deterministicka - jednou muze vypocet probehnout
# rychle, podruhe pomalu. Proto jsou testy rychlosti nekolikrat
# opakovany a namereny cas neni dobou behu jednoho vypoctu.
# I tak je cas pouze orientacni. Testy generuji grafy.


# TODO: Dopsat implementaci
def get_middle_element_of_array(array):
    """Funkce vrati prostredni prvek pole array. Pokud je delka pole
    suda, vrati se pravy prostredni prvek.
    Slozitost algoritmu by mela byt O(1).
    """
    pass


# TODO: Dopsat implementaci
def get_max_element_of_array(array):
    """Funkce vrati maximalni prvek pole array. Pokud je pole prazdne,
    vrati None.
    Slozitost algoritmu by mela byt O(n).
    """
    pass

# TODO: Dopsat implementaci
def minimum_difference_pair(array):
    """Funkce vrati dvojici cisel, ktera maji nejmensi rozdil. Poku je
    pole prazdne, vrati None.
    Slozitost algoritmu by mela byt O(n^2).
    """
    pass

# TODO: Dopsat implementaci
def generate_pin_combinations(n):
    """Funkce generuje PIN - kombinace vsech cisel do velkosti n.
    Slozitost algoritmu by mela byt O(10^n).
    """
    pass

# ---------------------------- #
#      Testy implementace      #
# ---------------------------- #
def test_get_middle_element_of_array():
    print("Test 1. get_middle_element_of_array, jednoduche testy")
    def simple_tests():
        assert get_middle_element_of_array([1, 2, 3, 4, 5]) == 3
        assert get_middle_element_of_array([1, 2, 3, 4, 5, 6]) == 3
        assert get_middle_element_of_array([1, 2, 3, 4, 5, 6, 7]) == 4

    simple_tests()  

    print("Test 1. get_middle_element_of_array, casova analyza")
    array_sizes = range(10, 10_000, 50)
    execution_times = []
    for size in array_sizes:
        array = random.sample(range(size), size)
        times = timeit.timeit(lambda: get_middle_element_of_array(array), number=100)
        execution_times.append(times)

    plt.title('Get Middle Element of Array')
    plt.plot(array_sizes, execution_times, 'or', label='Execution Time')
    
    plt.plot(array_sizes, np.poly1d(np.polyfit(array_sizes, execution_times, 1))(array_sizes), '-b', label='Linear Regression')
    plt.xlabel('Array Size')
    plt.ylabel('Execution Time (s)')
    plt.legend()
    plt.savefig('middle_element.png')
    plt.clf()

def test_get_max_element_of_array():
    print("Test 2. get_max_element_of_array, jednoduche testy")
    def simple_tests():
        assert get_middle_element_of_array([1, 2, 3, 4, 5]) == 3
        assert get_middle_element_of_array([1, 2, 3, 4, 5, 6]) == 3
        assert get_middle_element_of_array([1, 2, 3, 4, 5, 6, 7]) == 4

    simple_tests() 

    print("Test 2. get_max_element_of_array, casova analyza")
    array_sizes = range(10, 10_000, 50)
    execution_times = []
    execution_times_builtin = []
    for size in array_sizes:
        array = random.sample(range(size), size)
        times = timeit.timeit(lambda: get_max_element_of_array(array), number=100)
        times_builtin = timeit.timeit(lambda: max(array), number=100)
        execution_times.append(times)
        execution_times_builtin.append(times_builtin)

    plt.title('Get Max Element of Array')
    plt.plot(array_sizes, execution_times, 'or', label='Execution Time')
    plt.plot(array_sizes, execution_times_builtin, '-r', label='Builtin Execution Time')
    plt.plot(array_sizes, np.poly1d(np.polyfit(array_sizes, execution_times, 1))(array_sizes), '-b', label='Linear Regression')
    plt.xlabel('Array Size')
    plt.ylabel('Execution Time (s)')
    plt.legend()
    plt.savefig('max_element.png')
    plt.clf()

def test_minimum_difference_pair():
    print("Test 3. minimum_difference_pair, jednoduche testy")
    def simple_tests():
        assert minimum_difference_pair([1, 2, 3, 4, 5]) == (4, 5)
        assert minimum_difference_pair([1, 2, 3, 4, 5, 6]) == (5, 6)
        assert minimum_difference_pair([1, 2, 3, 4, 5, 6, 7]) == (6, 7)

    simple_tests()

    print("Test 3. minimum_difference_pair, casova analyza (moze trvat dlhsie)")
    array_sizes = range(10, 1000, 100)
    execution_times = []
    for size in array_sizes:
        array = random.sample(range(size), size)
        times = timeit.timeit(lambda: minimum_difference_pair(array), number=100)
        execution_times.append(times)

    plt.title('Minimum Difference Pair')
    plt.plot(array_sizes, execution_times, 'or', label='Execution Time')
    plt.plot(array_sizes, np.poly1d(np.polyfit(array_sizes, execution_times, 2))(array_sizes), '-b', label='Linear Regression')
    plt.xlabel('Array Size')
    plt.ylabel('Execution Time (s)')
    plt.legend()
    plt.savefig('smallest_difference.png')
    plt.clf()

def test_generate_pin_combinations():
    print("Test 4. generate_pin_combinations, jednoduche testy")
    def simple_tests():
        assert generate_pin_combinations(1) == [str(i) for i in range(10)]
        assert generate_pin_combinations(2) == [str(i) + str(j) for i in range(10) for j in range(10)]
        assert generate_pin_combinations(3) == [str(i) + str(j) + str(k) for i in range(10) for j in range(10) for k in range(10)]

    simple_tests()

    print("Test 4. generate_pin_combinations, casova analyza")
    pin_sizes = range(2, 7)
    execution_times = []
    for size in pin_sizes:
        times = timeit.timeit(lambda: generate_pin_combinations(size), number=10)
        execution_times.append(times)

    plt.title('PIN Combinations')
    plt.plot(pin_sizes, execution_times, 'or', label='Execution Time')
    plt.plot(pin_sizes, np.poly1d(np.polyfit(pin_sizes, execution_times, 5))(pin_sizes), '-r', label='Linear Regression')
    plt.xlabel('PIN Size')
    plt.ylabel('Execution Time (s)')
    plt.legend()
    plt.savefig('pin_combinations.png')
    plt.clf()


if __name__ == '__main__':
    test_get_middle_element_of_array()
    test_get_max_element_of_array()
    test_minimum_difference_pair()
    test_generate_pin_combinations()
