#!/usr/bin/env python3
import random
import timeit
import matplotlib.pyplot as plt
import numpy as np


# Your task in this implementation is to program the basic
# sorting algorithms. All algorithms will work with an array of numbers.
#
# You must implement the Bubble Sort, Selection Sort, MergeSort algorithms.


def swap(array, i, j):
    """Pomocna funkce swap pro vas, bere argumenty pole 'array', ve kterem
    prohodi prvky na pozicich indexu 'i' a 'j'.
    """
    array[i], array[j] = array[j], array[i]


def selection_sort(array):
    """Implement the Selection sort algorithm. Sort the given array `array`.
    The time complexity of Selection sort is O(n2) in both best and worst cases. 
    """
    pass

def bubble_sort(array):
    """Implement the Bubble sort. Sort the given array `array`.
    The time complexity of Bubble sort is O(n) in the best case and O(n2) in the worst case. 
    """
    pass

def merge(array, aux, left, mid, right):
    """Merging after sorting by MergeSort. Specified array 'array'
    contains 2 arranged sequences in intervals from 'left' to 'mid'
    and from 'mid'+1 to 'right'. Use the auxiliary array 'aux' to merge.
    Store the result in the 'array' field.
    """
    pass


def merge_sort(array, aux, left, right):
    """Implement the MergeSort. Seradte zadane pole 'array'
    in the interval from the index 'left' to index 'right'.
    Use the auxiliary array 'aux' to merge.
    Store the result in the 'array' field.
    """
    pass

### BONUS ###
def bogo_sort(array):
    """Implement the Bogo sort algorithm. Sort the given array `array`.
    The time complexity of Bogo sort is O(n!) in the average case. 
    """
    while not is_correct_result(array, sorted(array)):
        random.shuffle(array)
    return array

# Code for tests, do not edit.

ELEMENT_COUNT = 10
MAX_ELEMENT = 50


def new_random_array(size, max_element=MAX_ELEMENT):
    """Slouzi pro generovani noveho pole delky size."""
    return [random.randint(1, max_element) for _ in range(size)]


def is_correct_result(in_array, out_array):
    return sorted(in_array) == out_array


def run_merge_sort(array):
    aux = [0 for i in range(len(array))]
    merge_sort(array, aux, 0, len(array) - 1)


def test_sort(sort, MAX_ORDER=3):
    print("Test 1.: jednoprvkove pole [1]: ")
    array1 = [1]
    orig = array1[:]
    sort(array1)

    if is_correct_result(orig, array1):
        print("OK")
    else:
        print("NOK, puvodni pole bylo [1] a bylo serazeno,")
        print("po volani sortu je vystup: {}".format(array1[0]))

    print("Test 2.: dvouprvkove pole [1, 2]: ")
    array2 = [1, 2]
    orig = array2[:]
    sort(array2)

    if is_correct_result(orig, array2):
        print("OK")
    else:
        print("NOK, puvodni pole bylo [1, 2] a bylo serazeno,")
        print("po volani sortu je vystup: {}".format(array2))

    print("Test 3.: dvouprvkove pole [2, 1]: ")
    array3 = [2, 1]
    orig = array3[:]
    sort(array3)

    if is_correct_result(orig, array3):
        print("OK")
    else:
        print("NOK, puvodni pole bylo [2, 1],")
        print("po volani sortu je vystup: {}".format(array3))

    print("Test 4.: serazene pole 10 prvku [1..10]: ")
    array4 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    orig = array4[:]
    sort(array4)

    if is_correct_result(orig, array4):
        print("OK")
    else:
        print("NOK, puvodni pole bylo [1, 2..10],")
        print(" po volani sortu je vystup: {}".format(array4))

    print("Test 5.: neserazene pole: ")
    array5 = new_random_array(ELEMENT_COUNT)
    print("Puvodni pole: {}".format(array5))
    orig = array5[:]
    sort(array5)

    if is_correct_result(orig, array5):
        print("OK")
    else:
        print("NOK: po volani sortu je vystup: {}".format(array5))

    print("Test 6.: vykonnostni test na pole o velikosti 1, 10, ..., 10^MAX_SIZE:")
    execution_times = []
    SIZES = [i for i in range(1, 10**MAX_ORDER, 50)]
    print("Test 6.1: Random pole, max pocet prvku: {}".format(10**MAX_ORDER))
    for size in SIZES:
        array6 = new_random_array(size)
        times = timeit.timeit(lambda: sort(array6), number=10)
        execution_times.append(times)
        array6 = None

    execution_times_sorted = []
    print("Test 6.2: Sorted pole, max pocet prvku: {}".format(10**MAX_ORDER))
    for size in SIZES:
        array7 = [i for i in range(1, size)]
        times = timeit.timeit(lambda: sort(array7), number=10)
        execution_times_sorted.append(times)
        array7 = None

    plt.title('Sort Benchmark '+sort.__name__)
    plt.plot(SIZES, execution_times, 'or', label='Execution Time Random')
    plt.plot(SIZES, np.poly1d(np.polyfit(SIZES, execution_times, 2))(SIZES), '-r', label='Linear Regression Random')
    plt.plot(SIZES, execution_times_sorted, 'or', label='Execution Time Sorted')
    plt.plot(SIZES, np.poly1d(np.polyfit(SIZES, execution_times_sorted, 2))(SIZES), '-b', label='Linear Regression Sorted')
    plt.xlabel('Array Size')
    plt.ylabel('Execution Time (s)')
    plt.legend()
    
    plt.savefig('sort_benchmark_'+sort.__name__+'.png')
    plt.clf()

def test_merge():
    print("\nTesty na Merge z MergeSortu:\n")
    aux = [0 for i in range(ELEMENT_COUNT)]

    print("Test 1.: Merge 2 jednoprvkovych stejnych poli [1] a [1]:")
    array1 = [1, 1]
    merge(array1, aux, 0, 0, 1)

    if array1[0] == 1 and array1[0] == 1:
        print("OK")
    else:
        print("NOK, puvodni pole bylo [1, 1],")
        print("po volani Merge je vystup: {}".format(array1))

    print("Test 2.: Merge 2 jednoprvkovych stejnych poli [1] a [2]:")
    array2 = [1, 2]
    merge(array2, aux, 0, 0, 1)

    if array2[0] == 1 and array2[1] == 2:
        print("OK")
    else:
        print("NOK, puvodni pole bylo [1, 2],")
        print("po volani Merge je vystup: {}".format(array2))

    print("Test 3.: dvouprvkove pole [2, 1]: ")
    array3 = [2, 1]
    merge(array3, aux, 0, 0, 1)

    if array3[0] == 1 and array3[1] == 2:
        print("OK")
    else:
        print("NOK, puvodni pole bylo [2, 1], vysledek mel byt [1, 2],")
        print("po volani Merge je vystup: {}".format(array3))

    print("Test 4.: desetiprvkove pole [1, 3, 5, 7, 9, 2, 4, 6, 8, 10]:")
    array4 = [1, 3, 5, 7, 9, 2, 4, 6, 8, 10]
    orig = array4[:]
    merge(array4, aux, 0, 4, 9)

    if is_correct_result(orig, array4):
        print("OK")
    else:
        print("NOK, vysledek mel byt [1..10],")
        print("po volani Merge je vystup: {}".format(array4))

    print("Test 5.: desetiprvkove pole [1, 1, 2, 2, 3, 1, 1, 2, 3, 3]:")
    array5 = [1, 1, 2, 2, 3, 1, 1, 2, 3, 3]
    orig = array5[:]
    merge(array5, aux, 0, 4, 9)

    if is_correct_result(orig, array5):
        print("OK")
    else:
        print("NOK, vysledek mel byt [1, 1, 1, 1, 2, 2, 2, 3, 3, 3],")
        print("po volani Merge je vystup: {}".format(array5))


def test_merge_sort():
    print("\nTesty na MergeSort:")
    test_sort(run_merge_sort)


def test_bubble_sort():
    print("\nTesty na BubbleSort:")
    test_sort(bubble_sort)

def test_selection_sort():
    print("\nTesty na SelectionSort:")
    test_sort(selection_sort)

def test_bogo_sort():
    print("\nTesty na Bogo Sort:")
    test_sort(bogo_sort)



if __name__ == '__main__':
    test_selection_sort()
    test_bubble_sort()
    test_merge()
    test_merge_sort()
    test_bogo_sort()
