#!/usr/bin/env python3

class Node:
	"""Trida Node slouzi k reprezentaci uzlu ve strome.

	Atributy:
		key     klic daneho uzlu
		parent  reference na rodice uzlu (None, pokud neexistuje)
		left    reference na leveho potomka (None, pokud neexistuje)
		right   reference na praveho potomka (None, pokud neexistuje)
	"""

	def __init__(self):
		self.key = None
		self.parent = None
		self.right = None
		self.left = None


class BinarySearchTree:
	"""Trida BinarySearchTree slouzi k reprezentaci binarniho vyhledavaciho stromu.

	Atributy:
		root    reference na korenovy uzel typu Node
	"""

	def __init__(self):
		self.root = None


def height(tree):
	"""Vypocet vysky stromu. Vyska stromu je definovana jako delka nejdelsi cesty (pocet hran) od korene k listu.

	Vstup:
	tree    reference na instanci tridy BinarySearchTree

	Vystup:
	int     vyska stromu

	Pomocka: Pro vypocet maxima dvou cisel muzete pouzit funkci max(a, b). Vyuzijte pomocnou funkci pro rekurzivni pruchod stromem.
	"""
	pass


def search(tree, key):
	"""Vyhledani uzlu podle klice v binarnim vyhledavacim strome.

	Vstupy:
	tree    reference na instanci tridy BinarySearchTree
	key     hodnota klice

	Vystup:
	Node    reference na uzel s hodnotou klice, nebo None, pokud uzel s klicem neexistuje
	"""
	pass


def is_correct_bst(tree):
	"""Overi, zdali je strom 'tree' korektni binarni vyhledavaci strom.
	Pokud ano, vraci True, jinak False.
	
	Vstup:
	tree    reference na instanci tridy BinarySearchTree
	
	Vystup:
	bool    True, pokud je strom korektni binarni vyhledavaci strom, jinak False
	"""
	pass


def in_order_traversal(tree):
	"""Implementace funkce pro pruchod stromem metodou inorder.
	Vypise hodnoty uzlu ve strome setridene vzestupne.

	Vstup:
	tree    reference na instanci tridy BinarySearchTree
	
	Vystup:
	Array	pole hodnot uzlu ve strome setridene vzestupne
	
	Priklad:
	
		 5
		/ \
	   3   8
	  / \
	 1   4
	
	Vystup: 1, 3, 4, 5, 8

	Pomucka: Vytvorte pomocnou funkci pro rekurzivni pruchod stromem, ktera bude postupne pridavat hodnoty uzlu do pole.
	"""
	pass
	

def pre_order_traversal(tree):
	"""Implementace funkce pro pruchod stromem metodou preorder.

	Vstup:
	tree    reference na instanci tridy BinarySearchTree
	
	Vystup:
	Array	pole hodnot uzlu ve strome po preorder pruchodu

	Priklad:
	
		 5
		/ \
	   3   8
	  / \
	 1   4
	
	Vystup: 5, 3, 1, 4, 8

	Pomucka: Vytvorte pomocnou funkci pro rekurzivni pruchod stromem, ktera bude postupne pridavat hodnoty uzlu do pole.
	"""
	pass


def has_children_sum(tree):
	"""Funkce overi, zda pro kazdy uzel stromu plati, ze hodnota uzlu je rovna souctu hodnot jeho potomku.
	
	Vstup:
	tree    reference na instanci tridy BinarySearchTree
	
	Vystup:
	bool    True, pokud pro kazdy uzel stromu plati, ze hodnota uzlu je rovna souctu hodnot jeho potomku, jinak False
	"""
	pass

# Dodatek k graphvizu:
# Graphviz je nastroj, ktery vam umozni vizualizaci datovych struktur,
# coz se hodi predevsim pro ladeni. Tento program generuje nekolik
# souboru neco.dot v mainu. Vygenerovane soubory nahrajte do online
# nastroje pro zobrazeni graphvizu:
# http://sandbox.kidstrythisathome.com/erdos/
# nebo http://www.webgraphviz.com/- zvlada i vetsi grafy.
#
# Alternativne si muzete nainstalovat prekladac z jazyka dot do obrazku
# na svuj pocitac.
def make_graphviz(node, dot_file):
	if node is None:
		return

	dot_file.write('"{}" [label="{}"]\n'.format(id(node), node.key))

	for child, side in (node.left, 'L'), (node.right, 'R'):
		if child is not None:
			dot_file.write('"{}" -> "{}"\n'.format(id(node), id(child)))
			make_graphviz(child, dot_file)
		else:
			dot_file.write('{nil} [label="",color=white]\n{node} -> {nil}\n'
						   .format(node=id(node), nil=side + str(id(node))))


def make_graph(tree, filename):
	try:
		with open(filename, 'w') as dot_file:
			dot_file.write("digraph Tree {\n")
			dot_file.write("node [color=lightblue2, style=filled];\n")
			if tree is not None and tree.root is not None:
				make_graphviz(tree.root, dot_file)
			dot_file.write("}\n")
		print("Vykresleny strom najdete v souboru", filename)
	except Exception:
		print("Ve vykreslovani nastala chyba")

def init_test_tree():
	tree = BinarySearchTree()

	nodes = [Node() for _ in range(7)]
	for i in range(7):
		nodes[i].key = i

	tree.root = nodes[3]

	tree.root.left = nodes[1]
	nodes[1].parent = tree.root
	nodes[1].left = nodes[0]
	nodes[0].parent = nodes[1]
	nodes[1].right = nodes[2]
	nodes[2].parent = nodes[1]

	tree.root.right = nodes[5]
	nodes[5].parent = tree.root
	nodes[5].left = nodes[4]
	nodes[4].parent = nodes[5]
	nodes[5].right = nodes[6]
	nodes[6].parent = nodes[5]

	return tree


def helper_test_height(tree):
	tree_height = height(tree)

	if tree_height != 2:
		print("NOK - vyska 2 != vase vyska {}".format(tree_height))
		make_graph(tree, "height1.dot")
		return False

	node = Node()
	node.key = 7
	tree.root.right.right.right = node
	node.parent = tree.root.left.right

	tree_height = height(tree)

	if tree_height != 3:
		print("NOK - vyska 3 != vase vyska {}".format(tree_height))
		make_graph(tree, "height2.dot")
		return False

	print("OK")
	return True


def test_height():
	print("Test 1. height: ", end='')
	tree = init_test_tree()

	helper_test_height(tree)

def helper_test_search(tree):
	node = search(tree, 3)

	if node is None or node.key != 3:
		print("NOK - chybne hledani korene s hodnotou 3")
		return False

	node = search(tree, 2)

	if node is None or node.key != 2:
		print("NOK - chybne hledani listu s hodnotou 2")
		return False

	node = search(tree, 7)

	if node is not None:
		print("NOK - hledani prvku, ktery se ve strome nevyskytuje")
		return False

	print("OK")
	return True


def test_search():
	print("Test 2. search: ", end='')
	tree = init_test_tree()

	if not helper_test_search(tree):
		make_graph(tree, "search.dot")

def helper_test_is_correct_bst(tree):
	if not is_correct_bst(tree):
		print("NOK - strom je korektni binarni vyhledavaci strom")
		return False

	tree.root.key = 0
	tree.root.left.left = None

	if is_correct_bst(tree):
		print("NOK - strom neni korektni binarni vyhledavaci strom")
		return False

	tree.root.key = 3
	tree.root.left.right.key = 4
	tree.root.right.left.key = 2

	if is_correct_bst(tree):
		print("NOK - strom neni korektni binarni vyhledavaci strom")
		return False

	print("OK")
	return True


def test_is_correct_bst():
	print("Test 3. is_correct_bst: ", end='')
	tree = init_test_tree()

	if not helper_test_is_correct_bst(tree):
		make_graph(tree, "correct.dot")

def test_in_order_traversal():
	print("Test 4. in_order_traversal: ", end='')

	tree = BinarySearchTree()
	tree.root = Node()
	tree.root.key = 1
	tree.root.left = Node()
	tree.root.left.key = 2

	if not in_order_traversal(tree) == [2, 1]:
		make_graph(tree, "in_order1.dot")
		print("NOK")

	tree = init_test_tree()

	if not in_order_traversal(tree) == [0, 1, 2, 3, 4, 5, 6]:
		make_graph(tree, "in_order2.dot")
		print("NOK")

	print("OK")

def test_pre_order_traversal():
	print("Test 5. pre_order_traversal: ", end='')
	tree = init_test_tree()

	if pre_order_traversal(tree.root) == [3, 1, 0, 2, 5, 4, 6]:
		print("OK")
	else:
		print("NOK")

def test_has_children_sum():
	print("Test 6. has_children_sum: ", end='')
	tree = init_test_tree()

	if has_children_sum(tree):
		make_graph(tree, "has_sum_1.dot")
		print("NOK")

	tree.root.key = 1

	if has_children_sum(tree):
		make_graph(tree, "has_sum_2.dot")
		print("NOK")

	tree.root.key = 12
	tree.root.left.key = 2
	tree.root.right.key = 10

	if not has_children_sum(tree):
		make_graph(tree, "has_sum_3.dot")
		print("NOK")

	tree = BinarySearchTree()
	tree.root = Node()
	tree.root.key = 10
	
	if not has_children_sum(tree):
		make_graph(tree, "has_sum_3.dot")
		print("NOK")

	print("OK")

if __name__ == '__main__':
	test_search()
	test_height()
	test_is_correct_bst()
	test_in_order_traversal()
	test_pre_order_traversal()
	test_has_children_sum()

